/*
 *   Copyright 2017-2020, Schlumberger
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.opengroup.osdu.notification.api;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Strings;
import com.google.gson.Gson;
import com.google.gson.JsonElement;
import com.google.gson.JsonParser;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.cryptographic.ISignatureService;
import org.opengroup.osdu.core.common.http.HttpClient;
import org.opengroup.osdu.core.common.http.HttpRequest;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.notification.*;
import org.opengroup.osdu.core.common.notification.ISubscriptionFactory;
import org.opengroup.osdu.core.common.notification.ISubscriptionService;
import org.opengroup.osdu.core.common.notification.SubscriptionException;
import org.opengroup.osdu.notification.di.SubscriptionCacheFactory;
import org.opengroup.osdu.notification.provider.interfaces.IPubsubHandshakeHandler;
import org.opengroup.osdu.notification.provider.interfaces.IPubsubRequestBodyExtractor;
import org.opengroup.osdu.notification.utils.Config;
import org.opengroup.osdu.notification.provider.interfaces.IGoogleServiceAccount;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.context.annotation.RequestScope;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


@RestController
@RequestScope
@RequestMapping("/push-handlers")
public class PubsubEndpoint {
    @Autowired
    private IPubsubRequestBodyExtractor pubsubRequestBodyExtractor;
    @Autowired
    private IPubsubHandshakeHandler pubsubHandshakeHandler;
    @Autowired
    private ISignatureService signatureService;
    @Autowired
    private HttpClient httpClient;
    @Autowired
    private IGoogleServiceAccount gsaTokenProvider;
    @Autowired
    private JaxRsDpsLog log;
    @Autowired
    private SubscriptionCacheFactory subscriptionCacheFactory;
    @Autowired
    private ISubscriptionFactory subscriptionFactory;
    @Autowired
    private DpsHeaders headers;

    private static final String HMAC_TYPE = "HMAC";
    private static final String GSA_TYPE = "GSA";
    private final int WAITING_TIME = 30000;
    private final String ACKNOWLEDGE = "message acknowledged by client";
    private final String NOT_ACKNOWLEDGE = "message not acknowledged by client";
    private static final Gson gson = new Gson();
    private ObjectMapper objectMapper;

    @PostMapping("/records-changed")
    @PreAuthorize("@authorizationFilter.hasAnyPermission('" + Config.OPS + "', '" + Config.PUBSUB + "')")
    public ResponseEntity recordChanged() throws Exception {
        try {
            if (this.pubsubRequestBodyExtractor.isHandshakeRequest()) {
                String handshakeResponse = this.pubsubHandshakeHandler.getHandshakeResponse();
                return ResponseEntity.ok(handshakeResponse);
            }

            String notificationId = this.pubsubRequestBodyExtractor.extractNotificationIdFromRequestBody();
            String pubsubMessage = this.pubsubRequestBodyExtractor.extractDataFromRequestBody();
            Map<String, String> headerAttributes = this.pubsubRequestBodyExtractor.extractAttributesFromRequestBody();

            Subscription subscription = getSubscriptionFromCache(notificationId);
            Secret secret = subscription.getSecret();
            String endpoint = subscription.getPushEndpoint();

            String secretType = secret.getSecretType();
            String pushUrl = "";
            Map<String, String> requestHeader = new HashMap<>();

            if (secretType.equalsIgnoreCase(HMAC_TYPE)) {
                this.log.info("receiving pubsub message, will send out hmac type request, pubsub message: " + pubsubMessage);
                HmacSecret hmacSecret = (HmacSecret) secret;
                String signedjwt = this.signatureService.getSignedSignature(endpoint, hmacSecret.getValue());
                pushUrl = endpoint + "?hmac=" + signedjwt;
            } else if (secretType.equalsIgnoreCase(GSA_TYPE)) {
                this.log.info("receiving pubsub message, will send out gsa type request, pubsub message: " + pubsubMessage);
                GsaSecret gsaSecret = (GsaSecret) secret;
                GsaSecretValue gsaSecretValue = gsaSecret.getValue();

                JsonParser jsonParser = new JsonParser();
                JsonElement root = jsonParser.parse(gsaSecretValue.getKey());
                String keyString = root.getAsJsonObject().toString();

                String idToken = this.gsaTokenProvider.getIdToken(keyString, gsaSecretValue.getAudience());
                pushUrl = endpoint;
                requestHeader.put("Authorization", idToken);
            }

            this.log.info("sending out notification to endpoint: " + endpoint);
            requestHeader.put(DpsHeaders.CONTENT_TYPE, "application/json");
            requestHeader.put(DpsHeaders.CORRELATION_ID, headerAttributes.get(DpsHeaders.CORRELATION_ID));
            requestHeader.put(DpsHeaders.DATA_PARTITION_ID, headerAttributes.get(DpsHeaders.DATA_PARTITION_ID));

            HttpRequest request = HttpRequest.post().url(pushUrl).headers(requestHeader).body(pubsubMessage).connectionTimeout(WAITING_TIME).build();
            HttpResponse response = httpClient.send(request);
            if (!response.isSuccessCode()) {
                this.log.error(NOT_ACKNOWLEDGE);
                return ResponseEntity.badRequest().body(NOT_ACKNOWLEDGE);
            }
            this.log.info(ACKNOWLEDGE);
            return ResponseEntity.ok(ACKNOWLEDGE);
        } catch (Exception e) {
            e.printStackTrace();
        }
        return ResponseEntity.badRequest().body(NOT_ACKNOWLEDGE);
    }

    private Subscription getSubscriptionFromCache(String notificationId) throws Exception {
        String subscriptionString = subscriptionCacheFactory.get(notificationId);
        try {
            if (Strings.isNullOrEmpty(subscriptionString))
                subscriptionString = querySubscriptionAndUpdateCache(notificationId);
            ObjectMapper objectMapper = this.getObjectMapper();
            Subscription subscription = objectMapper.readValue(subscriptionString, Subscription.class);
            return subscription;
        } catch (IOException e) {
            this.log.warning("Error Parsing subscription String to object.");
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Error in getting subscription for notificationId:" + notificationId, "Unexpected error in pushing message", e);
        } catch (SubscriptionException se) {
            this.log.warning("Error query subscription from registration.");
            throw new AppException(HttpStatus.SC_INTERNAL_SERVER_ERROR, "Error in getting subscription for notificationId:" + notificationId, "Unexpected error in pushing message", se);
        }
    }

    private String querySubscriptionAndUpdateCache(String notificationId) throws Exception {
        ISubscriptionService service = subscriptionFactory.create(headers);

        List<Subscription> subscriptionList = service.query(notificationId);
        if (subscriptionList == null || subscriptionList.size() == 0) {
            this.log.warning(String.format("Subscription with notification ID %s not found in registration", notificationId));
            throw new AppException(HttpStatus.SC_NOT_FOUND, "Not found subscription for notificationId:" + notificationId, "Subscription not found");
        }

        Subscription subscription = subscriptionList.get(0);
        String jsonSubscription = gson.toJson(subscription);
        this.subscriptionCacheFactory.put(subscription.getNotificationId(), jsonSubscription);

        return jsonSubscription;
    }

    //unit test purpose
    protected ObjectMapper getObjectMapper() {
        if (this.objectMapper == null) {
            this.objectMapper = new ObjectMapper();
        }
        return this.objectMapper;
    }

    //unit test purpose
    void setObjectMapper(ObjectMapper objectMapper) {
        this.objectMapper = objectMapper;
    }
}
