// Copyright © 2020 Amazon Web Services
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.notification.provider.aws.impl;

import com.amazonaws.auth.AWSCredentialsProvider;
import com.amazonaws.services.secretsmanager.AWSSecretsManager;
import com.amazonaws.services.secretsmanager.AWSSecretsManagerClientBuilder;
import com.amazonaws.services.secretsmanager.model.*;
import com.amazonaws.services.simplesystemsmanagement.AWSSimpleSystemsManagement;
import com.amazonaws.services.simplesystemsmanagement.AWSSimpleSystemsManagementClientBuilder;
import com.amazonaws.services.simplesystemsmanagement.model.*;
import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;

import org.opengroup.osdu.core.aws.iam.IAMConfig;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.util.IServiceAccountJwtClient;
import org.opengroup.osdu.notification.provider.aws.utils.AwsCognitoClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;

import java.io.IOException;

import java.util.ArrayList;

import java.util.List;
import java.util.Map;

@Component
public class ServiceAccountJwtAwsClientImpl implements IServiceAccountJwtClient {


    @Value("${aws.region}")
    @Getter()
    @Setter(AccessLevel.PROTECTED)
    public String amazonRegion;



    @Value("${aws.ssm}")
    @Getter()
    @Setter(AccessLevel.PROTECTED)
    public Boolean ssmEnabled;


    @Value("${aws.environment}")
    @Getter()
    @Setter(AccessLevel.PROTECTED)
    public String environment;


    @Autowired
    private JaxRsDpsLog log;

    String password;
    String clientid;
    String userpoolid;
    String serviceprincipaluser;
    AwsCognitoClient cognitoClient;

    private AWSCredentialsProvider amazonAWSCredentials;
    private AWSSimpleSystemsManagement ssmManager;

    @PostConstruct
    public void init() {
        if (ssmEnabled) {
            String secretKey = "service_principal_password";
            String secretName = "/osdu/" + environment + "/service_principal_password";
            String cognito_user_pool_id = "/osdu/" + environment + "/cognito-user-pool-id";
            String cognito_client_id = "/osdu/" + environment + "/cognito-client-id";
            String service_principal = "/osdu/" + environment + "/service-principal-user";
            amazonAWSCredentials = IAMConfig.amazonAWSCredentials();
            ssmManager = AWSSimpleSystemsManagementClientBuilder.standard()
                    .withCredentials(amazonAWSCredentials)
                    .withRegion(amazonRegion)
                    .build();

            GetParametersRequest paramRequest = new GetParametersRequest()
                    .withNames(cognito_user_pool_id,cognito_client_id,service_principal)
                    .withWithDecryption(true);
            GetParametersResult paramResult = new GetParametersResult();
            paramResult = ssmManager.getParameters(paramRequest);
            List<Parameter> paramsResultList = new ArrayList<>();
            List<String> paramsResultListInvalid = new ArrayList<>();
            paramsResultList = paramResult.getParameters();
            paramsResultListInvalid = paramResult.getInvalidParameters();

            if(paramsResultListInvalid.size() >0)
            {
                log.error("SSM did not retrieve all parameters");
            }
            for (Parameter s : paramsResultList) {
                if (s.getName().equalsIgnoreCase(cognito_user_pool_id)) {
                    userpoolid = s.getValue();
                }
                if (s.getName().equalsIgnoreCase(cognito_client_id)) {
                    clientid = s.getValue();
                }
                if (s.getName().equalsIgnoreCase(service_principal)) {
                    serviceprincipaluser = s.getValue();
                }

            }

            password = getSecret(secretName,amazonRegion,secretKey);
             cognitoClient = new AwsCognitoClient(amazonRegion,clientid,"USER_PASSWORD_AUTH", serviceprincipaluser,password);
             cognitoClient.setPassword(serviceprincipaluser,password,userpoolid);
        }
    }

    @Override
    public String getIdToken(String s) {

        String token=  getServicePrincipalCredentials();
        return token;

    }

    public  String getServicePrincipalCredentials()
    {

        String token = cognitoClient.getToken(serviceprincipaluser,password,"bearer");
        return token;

    }

    public  String getSecret(String secretName, String region,String secretKey) {


String secretVaue="";
        // Create a Secrets Manager client
        AWSSecretsManager client  = AWSSecretsManagerClientBuilder.standard()
                .withRegion(region)
                .build();

        String secret="", decodedBinarySecret="";
        GetSecretValueRequest getSecretValueRequest = new GetSecretValueRequest()
                .withSecretId(secretName);
        GetSecretValueResult getSecretValueResult = null;

        try {
            getSecretValueResult = client.getSecretValue(getSecretValueRequest);
        } catch (DecryptionFailureException e) {
            // Secrets Manager can't decrypt the protected secret text using the provided KMS key.
            // Deal with the exception here, and/or rethrow at your discretion.
            log.error("Error while setting up ServicePrincipalAccount"+e.getMessage());
            throw e;
        } catch (InternalServiceErrorException e) {
            // An error occurred on the server side.
            // Deal with the exception here, and/or rethrow at your discretion.
            log.error("Error while setting up ServicePrincipalAccount"+e.getMessage());
            throw e;
        } catch (InvalidParameterException e) {
            // You provided an invalid value for a parameter.
            // Deal with the exception here, and/or rethrow at your discretion.
            log.error("Error while setting up ServicePrincipalAccount"+e.getMessage());
            throw e;
        } catch (InvalidRequestException e) {
            // You provided a parameter value that is not valid for the current state of the resource.
            // Deal with the exception here, and/or rethrow at your discretion.
            log.error("Error while setting up ServicePrincipalAccount"+e.getMessage());
            throw e;
        } catch (ResourceNotFoundException e) {
            // We can't find the resource that you asked for.
            // Deal with the exception here, and/or rethrow at your discretion.
            log.error("Error while setting up ServicePrincipalAccount"+e.getMessage());
            throw e;
        }

        // Decrypts secret using the associated KMS CMK.
        // Depending on whether the secret is a string or binary, one of these fields will be populated.
        if (getSecretValueResult.getSecretString() != null) {
            secret = getSecretValueResult.getSecretString();
            Map<String, String> secretMap=null;

            try
            {
                secretMap = new ObjectMapper().readValue(secret.getBytes(), Map.class);

            } catch (JsonParseException e) {
                log.error(e.getMessage());
            } catch (JsonMappingException e) {
                log.error(e.getMessage());
            } catch (IOException e) {
                log.error(e.getMessage());
            }

             secretVaue = secretMap.get(secretKey);
        }


        return secretVaue;
    }


}
