// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.provider.azure.util;

import com.auth0.jwt.JWT;
import com.microsoft.aad.adal4j.AuthenticationContext;
import com.microsoft.aad.adal4j.AuthenticationResult;
import com.microsoft.aad.adal4j.ClientCredential;
import org.apache.http.HttpStatus;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.search.IdToken;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.provider.interfaces.IJwtCache;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.opengroup.osdu.core.common.util.IServiceAccountJwtClient;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.net.MalformedURLException;
import java.util.concurrent.*;

@Component
public class ServiceAccountJwtAzureClientImpl implements IServiceAccountJwtClient {

    @Autowired
    private AppProperties config;

    @Autowired
    private ITenantFactory tenantInfoServiceProvider;

    @Autowired
    private IJwtCache tenantJwtCache;

    public String getIdToken(String tenantName) {
        TenantInfo tenant = this.tenantInfoServiceProvider.getTenantInfo(tenantName);
        if (tenant == null) {
            throw new AppException(HttpStatus.SC_BAD_REQUEST, "Invalid tenant Name", "Invalid tenant Name from azure");
        }

        String ACCESS_TOKEN = "";
        ExecutorService service = null;

        try {
            // TODO : Refactor to move ID token form Common.Core.model.search to Common.core
            IdToken cachedToken = (IdToken) this.tenantJwtCache.get(tenant.getName());

            if ((cachedToken != null) && !IdToken.refreshToken(cachedToken)) {
                return "Bearer " + cachedToken.getTokenValue();
            }

            // TODO : Control the thread count via config and pool should be created once.
            service = Executors.newFixedThreadPool(1);

            ACCESS_TOKEN = getAccessToken(service);
            IdToken idToken = IdToken.builder().tokenValue(ACCESS_TOKEN).expirationTimeMillis(JWT.decode(ACCESS_TOKEN).getExpiresAt().getTime()).build();
            this.tenantJwtCache.put(tenant.getName(), idToken);
        }  finally {
            if(service != null) {
                service.shutdown();
            }
        }
        return "Bearer " + ACCESS_TOKEN;
    }

    // TODO : Refactor for making it test-able.
    // THIS METHOD IS PUBLIC ONLY TO ENABLE UNIT TESTING
    public String getAccessToken(ExecutorService service) {
        AuthenticationContext context = null;
        ClientCredential credential = null;
        String ACCESS_TOKEN = null;
        try {
            context = new AuthenticationContext(this.config.getAuthURL(), false, service);
            credential = new ClientCredential(this.config.getAuthClientID(), this.config.getAuthClientSecret());

            Future<AuthenticationResult> future = context.acquireToken(this.config.getAadClientID(), credential, null);

            if (future == null) {
                throw new AppException(HttpStatus.SC_FORBIDDEN, "Token not generated", "The user is not authorized to obtain Token From AAD");
            }
            ACCESS_TOKEN = future.get().getAccessToken();
        } catch (MalformedURLException malformedURLException) {
            malformedURLException.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (ExecutionException e) {
            e.printStackTrace();
        }
        return ACCESS_TOKEN;
    }
}

