// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.notification.util;

import org.apache.http.HttpStatus;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.mockito.junit.MockitoJUnitRunner;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.search.IdToken;
import org.opengroup.osdu.core.common.model.tenant.TenantInfo;
import org.opengroup.osdu.core.common.provider.interfaces.ITenantFactory;
import org.opengroup.osdu.notification.provider.azure.cache.JwtCache;
import org.opengroup.osdu.notification.provider.azure.util.AppProperties;
import org.opengroup.osdu.notification.provider.azure.util.ServiceAccountJwtAzureClientImpl;

import java.util.concurrent.ExecutorService;

import static org.junit.Assert.fail;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.MockitoAnnotations.initMocks;
import static org.mockito.Mockito.when;

@RunWith(MockitoJUnitRunner.class)
public class ServiceAccountClientImplTest {

    final String tenantName = "Test Tenant";
    final String validToken = "validToken";

    @Mock
    private ITenantFactory tenantInfoServiceProvider;

    @Mock
    private TenantInfo tenantInfo;

    @Mock
    private IdToken idToken;

    @Mock
    private ExecutorService executorService;

    @Mock
    private AppProperties appProperties;

    @Mock
    private JwtCache tenantJwtCacheMock;

    @Mock
    private JaxRsDpsLog logger;

    @InjectMocks
    @Spy
    private ServiceAccountJwtAzureClientImpl sut;

    @Before
    public void setup() {
        initMocks(this);
        when(tenantInfo.getName()).thenReturn(tenantName);
        when(this.tenantInfoServiceProvider.getTenantInfo(tenantName)).thenReturn(tenantInfo);
        idToken = IdToken.builder().tokenValue(validToken).expirationTimeMillis(System.currentTimeMillis() + 10000000L).build();
    }

    @Test
    public void should_throwAppException_getIdTokenTest() {
        // Set up
        when(this.tenantInfoServiceProvider.getTenantInfo(tenantName)).thenReturn(null);
        try {
            // Act
            sut.getIdToken(tenantName);

            // Assert
            fail("Should throw exception");
        } catch (AppException e) {
            Assert.assertEquals(HttpStatus.SC_BAD_REQUEST, e.getError().getCode());
            Assert.assertEquals("Invalid tenant Name from azure", e.getError().getMessage());
        } catch (Exception e) {
            fail("Should not throw this exception" + e.getMessage());
        }
    }

    @Test
    public void should_getTokenFromCache_getIdTokenTest() {
        // SetUp
        when(tenantJwtCacheMock.get(any())).thenReturn(idToken);
        String expectedToken = "Bearer " +idToken.getTokenValue();

        // Act
        String returnedIdToken = sut.getIdToken(tenantName);

        // Assert
        Assert.assertEquals(expectedToken, returnedIdToken);
    }

    @Test
    public void should_updateCache_getIdTokenTest() {
        // Set up
        when(tenantJwtCacheMock.get(any())).thenReturn(idToken);
        String expectedToken = "Bearer " +idToken.getTokenValue();

        // Act
        String returnedToken = this.sut.getIdToken(tenantName);

        // Assert
        Assert.assertEquals(expectedToken, returnedToken);
    }

    @Test
    public void should_return403GivenInvalidApplicationProperties_getAccessToken() {
        when(appProperties.getAuthURL()).thenReturn("https://login.microsoftonline.com/s/oauth2/token/");
        when(appProperties.getAuthClientID()).thenReturn("testAuthClientID");
        when(appProperties.getAuthClientSecret()).thenReturn("testAuthClientSecret");
        when(appProperties.getAadClientID()).thenReturn("testAadClientID");

        try {
            // Act
            sut.getAccessToken(executorService);

            // Assert
            fail("Should throw exception");
        } catch (AppException appException) {
            Assert.assertEquals(HttpStatus.SC_FORBIDDEN, appException.getError().getCode());
        } catch (Exception e) {
            fail("Should not throw this exception" + e.getMessage());
        }
    }
}

